/* $Id: ast_operatordeclaration.h 6310 2023-01-16 20:39:48Z aron $ */
/*
    Welltype parser and AST
    Copyright (C) 2014-2023  Aron Barath

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef WTPARSER_AST_OPERATORDECLARATION_H
#define WTPARSER_AST_OPERATORDECLARATION_H

#include "../astnode.h"
#include "ast_declaration.h"
#include "../../wtruntime/wtruntime.h"
#include <string>

namespace wtparser
{
class ParserHelper;
} // wtparser

namespace wtast
{

class OperatorDeclaration : public Declaration
{
	std::string mOpName;
	bool        mIsLeftAssoc;
	wt_i32      mPrec;
	bool        mBuiltin;

public:
	OperatorDeclaration(const std::string & opname, bool is_left, wt_i32 prec)
		: Declaration()
		, mOpName(opname)
		, mIsLeftAssoc(is_left)
		, mPrec(prec)
		, mBuiltin(false)
	{
		// ok
	}

	~OperatorDeclaration()
	{
		// ok
	}

	void Accept(AstVisitor* vis);

	OperatorDeclaration* Copy();

	Kind GetKind() const
	{
		return OPERATORDECLARATION;
	}

	const std::string & GetOpName() const
	{
		return mOpName;
	}

	bool IsLeftAssociative() const
	{
		return mIsLeftAssoc;
	}

	wt_i32 GetPrecedence() const
	{
		return mPrec;
	}

	const std::string & GetAssociativity() const;

	bool IsBuiltin() const
	{
		return mBuiltin;
	}

	friend class wtparser::ParserHelper;

private:

	void SetBuiltin()
	{
		mBuiltin = true;
	}
};

} // namespace wtast

#endif // WTPARSER_AST_OPERATORDECLARATION_H
