#ifndef WTRUNTIME_DYNSTACK_H
#define WTRUNTIME_DYNSTACK_H

#include "config.h"
#include "types.h"

struct wtr_profile;
struct wtr_objentry;

struct wtr_dynstack
{
	wt_u32               sp;       /* stack pointer */
	wt_u32               size;     /* stack size */
	wt_u32               min_size; /* minimal size of the stack (cannot be decreased below this value) */
	wt_u32               max_size; /* maximal size of the stack (cannot be increased over this value) */
	wt_u32               limit;    /* limit when the stack capacity will be decreased */
	struct wtr_objentry* data;     /* allocated data */
	struct wtr_profile*  profile;  /* the dynamic stack belongs to this virtual machine (where the allocations are made from) */
};

/* Initialize a dynamic stack. The shell structure must be preallocated. */
/* The stack size will be dynamically increased, up to 'maxsize'. */
/* In case of increase, the size will be doubles, as 'limit' as well. */
/* If the size reaches the actual value of 'limit', the capacity will */
/* be halved, so as 'limit'. The capacity cannot be smaller than 'minsize'. */
/* The initial capacity is 'minsize'. */
wt_bool wtr_dynstack_init(struct wtr_dynstack* const stk,
	struct wtr_profile* const prog, wt_u32 minsize, wt_u32 maxsize,
	wt_u32 limit);

/* Delete the stack. */
void wtr_dynstack_free(struct wtr_dynstack* const stk);

/* Store an element in the stack. Returns boolean indicating the success. */
wt_bool wtr_dynstack_push(struct wtr_dynstack* const stk,
	struct wtr_objentry* const entry);

/* Returns the address of the top object entry. */
/* Returns NULL if the stack is empty. */
struct wtr_objentry* wtr_dynstack_top(struct wtr_dynstack* const stk);

/* Removes the object entry on the top of the stack. */
/* Returns false if the stack was empty, true otherwise. */
wt_bool wtr_dynstack_pop(struct wtr_dynstack* const stk);

#endif /* WTRUNTIME_DYNSTACK_H */
