#ifndef WTRUNTIME_FRAME_H
#define WTRUNTIME_FRAME_H

#include "config_private.h"
#include "types.h"

#define WTRCF_DVT              0x00000001u /* DVT is allocated */
#define WTRCF_CSOS             0x00000002u /* CSOCS is allocated */
#define WTRCF_START            0x00000010u /* start frame, context shall enter to init state when this frame exits */

struct wtr_program;
struct wtr_thread;
struct wtr_objentry;

struct wtr_callframe
{
	wt_u32               pc;        /* actual Program Counter */
	wt_u32               flags;     /* callframe flags (WTRCF_*) */
	wt_u32               ehandler;  /* Exception handler */
	wt_u32               dvt;       /* Dynamic Variable Table offset from 'memory' */
	wt_u32               csos_sp;   /* CSOS Stack Pointer */
	wt_u32               tos_mark;  /* Temporary Object Stack mark */
	struct wtr_objentry* csos;      /* CallStack Object Stack */
	wt_u8*               dsp_mark;  /* DataStackPointer mark */
	struct wtr_program*  prog;      /* actual context */
};
typedef struct _wtr_callframe wtr_callframe;


/*

	F: Free
	T: stack Top
	R: Routine base data (rutinonkent fix meretu)
	N: Not visible (rutin szamara lathatatlan)

	F +-----------+  <-- Stack Top
	F |           |
	F |           |
	T |-----------|
	R | CSOS SIZE |  <-- FIX (CSOSALLOC, CSOSFREE, CSOSPOP, CSOSPUSH)
	R |-----------|
	R |           |
	R |  .  .  .  |
	R |           |
	R |-----------|
	R | DVT  SIZE |  <-- FIX (DVTALLOC, DVTFREE)
	R |-----------|
	R |           |
	R |  .  .  .  |
	R |           |
	N |-----------|
	N | CallFrame |
	N |-----------|
	N |           |
	N |  .  .  .  |
	N |           |
	N +-----------+  <-- Stack Bottom

*/

#endif /* WTRUNTIME_FRAME_H */
