#ifndef WTRUNTIME_PROFILE_H
#define WTRUNTIME_PROFILE_H

#include "config.h"
#include "types.h"
#include "wtruntime.h"
#include "vector.h"
#include "libvmma/libvmma.h"
#include <stddef.h>

struct wtr_profile;
struct wtr_program;
struct wtr_exception;
struct wtr_funcsig;
struct wtr_function;
struct wtr_type;
struct wtr_operator;

struct wtr_profile
{
	struct vmma_context*   alloc_context;   /* VMMA context */
	struct vmma_client*    alloc_client;    /* VMMA client, used to allocate profile-related memory */
	wt_u8*                 memory;          /* start address of the given or the allocated memory area */
	void*                  allocated_mem;   /* if 'memory' is allocated, this is the address returned by system memory allocator */
	wt_uintptr             total_memsize;   /* size of the 'memory' */
// TODO: this should be broken into 'num_cfunc' and 'num_fhan' OR make a list of those
	wt_u32                 refcount;        /* number of references to this profile */
	wt_u32                 num_context;     /* number of created context structures */
	wt_u32                 external_tag;    /* tag of the wt_pfn that holds an external C function */
	wt_u32                 accepted_flags;  /* flags accepted during the initialization (WTRPF_*) */
	wt_u32                 pc_tcl_ja;       /* absolute PC of the "TCL ja.s" routine */
	wt_u32                 pc_tcl_leave;    /* absolute PC of the context cleanup ("TCL leave") routine */
	wt_u32                 pc_strd_leave;   /* absolute PC of the "STRD leave" routine */
	wt_u32                 pc_cfp_leave;    /* absolute PC of the "CFP leave" routine */
	wt_u32                 ratio_css;       /* CSS ratio from 'stack_size' */
	wt_u32                 ratio_dss;       /* DSS ratio from 'stack_size' */
	void*                  alloc_ctx;       /* memory allocator context */
	WT_PFN_ALLOC           alloc_func;      /* memory allocator function */
	void*                  user_ptr;        /* an additional user pointer, this is opaque to the VM */
	struct wtr_list        list_programs;   /* list of all loaded programs */
	struct wtr_vector      exception_vec;   /* exceptions (global EXID -> exception struct) */
	struct wtr_list        list_functions;  /* list of wtr_function */
	struct wtr_list        list_operators;  /* list of wtr_operator */
	struct wtr_vector      type_vec;        /* installed types */
	wt_u32                 xs_x;            /* xorshift prng state */
	wt_u32                 xs_y;            /* xorshift prng state */
	wt_u32                 xs_z;            /* xorshift prng state */
	wt_u32                 xs_w;            /* xorshift prng state */
	void*                  loaderops_ctx;   /* unresolved object handlers */
	const struct wtr_loaderops* loaderops;  /* unresolved object handlers */
};

wt_error wtr_profile_addexception(struct wtr_profile* const vm,
	struct wtr_program* const prog, const char* const excname,
	struct wtr_exception** const p_exc);

struct wtr_exception* wtr_profile_findexception(struct wtr_profile* const vm,
	const char* const excname);

/* frees memory up */
wt_bool wtr_profile_removeexception(struct wtr_profile* const vm,
	struct wtr_exception* const exc);

struct wtr_function* wtr_profile_findfunction(struct wtr_profile* const vm,
	struct wtr_funcsig* const sig);

struct wtr_type* wtr_profile_findtype(struct wtr_profile* const vm,
	const char* const name);

wt_bool wtr_profile_addtype(struct wtr_profile* const vm,
	struct wtr_type* const type);

/* frees memory up */
void wtr_profile_removetype(struct wtr_profile* const vm,
	struct wtr_type* const type);

struct wtr_operator* wtr_profile_findoperator(struct wtr_profile* const vm,
	const char* const name);

wt_bool wtr_profile_addoperator(struct wtr_profile* const vm,
	struct wtr_operator* const op);

/* frees memory up */
void wtr_profile_removeoperator(struct wtr_profile* const vm,
	struct wtr_operator* constop);


/* determine the size of a WTT_* type (may have additional WTTF_*) */
size_t wtr_sizeof(wt_u32 type);


/* allocate memory that is accessible in the VM */
void* wtr_profile_vm_malloc(struct wtr_profile* prof, size_t size);
void  wtr_profile_vm_free  (struct wtr_profile* prof, void* memory);

/* reallocate system memory */
void* wtr_profile_sys_realloc(struct wtr_profile* const prof, void* memory,
	const size_t old_size, const size_t new_size);
void* wtr_profile_sys_malloc(struct wtr_profile* const prof, size_t size);
void wtr_profile_sys_free(struct wtr_profile* const prof, void* memory);

#endif /* WTRUNTIME_PROFILE_H */
