#ifndef WTRUNTIME_PUBLIC_H
#define WTRUNTIME_PUBLIC_H

#include "config.h"
#include <stddef.h>

#ifndef WTAPIENTRY
#  define WTAPIENTRY
#endif /* WTAPIENTRY */


/*----------------------------------------------------------------*/
/* version numbers */

#define WT_INSTRVER_BEHEMOTH    1
#define WT_INSTRVER_BEHEMOTH2   2
#define WT_INSTRVER_TOMB        3

#define WT_MAKEVERSION(_major_,_minor_,_patch_) \
	( \
		(((wt_u32)(_major_)) << 24) | \
		(((wt_u32)(_minor_)) << 16) | \
		(((wt_u32)(_patch_)) <<  8) \
	)

#define WT_GETMAJOR(_version_)  (((_version_)>>24) & 0xff)
#define WT_GETMINOR(_version_)  (((_version_)>>16) & 0xff)
#define WT_GETPATCH(_version_)  (((_version_)>> 8) & 0xff)


/*----------------------------------------------------------------*/
/* virtual machine flags */

#define WTRVMF_NOFLAGS         0x00000000ul /* placeholder */
#define WTRVMF_REQINTEGRITY    0x00000001ul /* loader requires validated binaries (at least one INTEGRITY section must pass) */


/*----------------------------------------------------------------*/
/* additional flags for loading programs */

#define WTRLF_NOFLAGS          0x00000000ul /* placeholder */
#define WTRLF_REQINTEGRITY     0x00000001ul /* loader requires validated binaries (at least one INTEGRITY section must pass) */


/*----------------------------------------------------------------*/
/* types IDs, type masks, and type flags */

#define WTTF_REF               0x80000000ul
#define WTTF_MUTABLE           0x40000000ul /* function argument type modifier: able to modify object content */
#define WTTF_SEQ               0x20000000ul
#define WTTF_LOCAL             0x10000000ul /* type from the local type vector (not resolved type) */
#define WTTF_USERTYPE          0x00800000ul
#define WTTM_USERTYPE          0x00fffffful
#define WTTM_USERTYPEID        0x007ffffful
#define WTTM_BASICTYPE         0x0000001ful
#define WTTM_USEDBITS          0xf0fffffful
#define WTTM_UNUSEDBITS        (~WTTM_USEDBITS)
#define WTT_VOID               0
#define WTT_BOOL               1
#define WTT_I32                2
#define WTT_U32                3
#define WTT_I64                4
#define WTT_U64                5
#define WTT_F32                6
#define WTT_F64                7
#define WTT_CHAR               8
#define WTT_UNUSED9            9   /* preserved for WCHAR */
#define WTT_STRING             10
#define WTT_UNUSED11           11  /* preserved for WSTRING */
#define WTT_HANDLE             12
#define WTT_PROGRAM            13
#define WTT_UNUSED14           14
#define WTT_UNUSED15           15
#define WTT_FILE               16
#define WTT_DIR                17
#define WTT_SOCKET             18  /* currently unused */
#define WTT_I8                 19
#define WTT_U8                 20
#define WTT_I16                21
#define WTT_U16                22
#define WTT_UNUSED23           23
#define WTT_UNUSED24           24
#define WTT_UNUSED25           25
#define WTT_UNUSED26           26
#define WTT_UNUSED27           27
#define WTT_UNUSED28           28
#define WTT_UNUSED29           29
#define WTT_UNUSED30           30
#define WTT_UNUSED31           31


/*----------------------------------------------------------------*/
/* boolean values */

#define WT_TRUE                1
#define WT_FALSE               0


/*----------------------------------------------------------------*/
/* error codes */

#define WTE_OK                 0      /* successful */
#define WTE_BADARG             0x0001 /* bad argument */
#define WTE_LFAIL              0x0002 /* low-level or system-wide failure */
#define WTE_NOENT              0x0003 /* no such entity */
#define WTE_NOMEM              0x0004 /* not enough memory */
#define WTE_NOFIT              0x0005 /* data does not fit in the given buffer */
#define WTE_DUPENT             0x0006 /* entity duplication detected */
#define WTE_REFER              0x0007 /* reference counter is non-zero */
#define WTE_BINERR             0x0008 /* general binary error */
#define WTE_BADVER             0x0009 /* bad binary version or unsupported feature */
#define WTE_UNRES              0x0010 /* unresolved external symbol */
#define WTE_AMBIG              0x0011 /* ambiguous symbol */
#define WTE_BADEX              0x0012 /* bad external (e.g. multiple or bad operator) */
#define WTE_URTEX              0x0013 /* unhandled runtime exception */
#define WTE_INVOKE             0x0014 /* failed to invoke main{} block or function */
#define WTE_INTEG              0x0015 /* integrity check failed */
#define WTE_MEMSZ              0x0016 /* specified memory is too small */
#define WTE_REFCTX             0x0017 /* number of context references is non-zero */
#define WTE_BADCTX             0x0018 /* context in bad state */
#define WTE_NOSTK              0x0019 /* stack is too small */
#define WTE_DESTRP             0x0020 /* program destructor invoked */
#define WTE_UCLEAN             0x0021 /* structure needs cleaning */
#define WTE_BADMAG             0x0022 /* bad magic number */
#define WTE_ABIVER             0x0023 /* binary ABI version out of supported range */


/*----------------------------------------------------------------*/
/* metadata keys */

#define WT_KEY_AUTHOR          ((const char*)"Author")
#define WT_KEY_TITLE           ((const char*)"Title")
#define WT_KEY_VERSION         ((const char*)"Version")
#define WT_KEY_COPYRIGHT       ((const char*)"Copyright")
#define WT_KEY_ID              ((const char*)"Id")
#define WT_KEY_COMMENT         ((const char*)"Comment")


/*----------------------------------------------------------------*/
/* invocation completions */

#define WTIC_INVOKED           0
#define WTIC_SUCCESS           1
#define WTIC_EXCEPTION         2


/*----------------------------------------------------------------*/
/* directory entry types */

#define WTDT_DIR               'D'
#define WTDT_REG               'R'
#define WTDT_BLK               'B'
#define WTDT_CHR               'C'
#define WTDT_LNK               'L'
#define WTDT_FIFO              'F'
#define WTDT_SOCK              'S'
#define WTDT_UNKNOWN           '?' /* entry type could not be determined */


/*----------------------------------------------------------------*/
/* file access modes */

#define WTFA_EXIST             0
#define WTFA_READ              4
#define WTFA_WRITE             2
#define WTFA_EXECUTE           1


/*----------------------------------------------------------------*/
/* file seek whence values */

#define WTSW_SET               1
#define WTSW_CUR               2
#define WTSW_END               3


/*----------------------------------------------------------------*/
/* types */

typedef wt_u32 wt_error;     /* holds an error code (WTE_*) */
typedef wt_u32 wt_bool;      /* WT_TRUE or WT_FALSE */
typedef wt_u32 wt_char;
typedef wt_u32 wt_exception;
typedef wt_u64 wt_handle;

typedef char* wt_string;


/*----------------------------------------------------------------*/
/* binary types */

#define WTBMAG0                'W'
#define WTBMAG1                'T'
#define WTBMAG2                'B'
#define WTBMAG3                '\x03'

#define WTBV_CURRENT           1

#define WTBO_LITTLE            1 /* little endian */
#define WTBO_BIG               2 /* big endian */

#define WTB_NIDENT             16 /* size of 'b_ident' */

#define WTBI_MAGIC0            0  /* value is WTBMAG0 */
#define WTBI_MAGIC1            1  /* value is WTBMAG1 */
#define WTBI_MAGIC2            2  /* value is WTBMAG2 */
#define WTBI_MAGIC3            3  /* value is WTBMAG3 */
#define WTBI_VERSION           4  /* WTBV_CURRENT */
#define WTBI_BYTEORDER         5  /* byte order (WTBO_*) */
/*#define WTBI_                6  */ /* MBZ */
/*#define WTBI_                7  */ /* MBZ */
#define WTBI_COMPVER0          8  /* compiler version (major) */
#define WTBI_COMPVER1          9  /* compiler version (minor) */
#define WTBI_COMPVER2          10 /* compiler version (patch) */
#define WTBI_COMPVER3          11 /* MBZ since v0.9.5; used to be 4th component of the compiler version */
/*#define WTBI_                12 */ /* MBZ */
/*#define WTBI_                13 */ /* MBZ */
/*#define WTBI_                14 */ /* MBZ */
/*#define WTBI_                15 */ /* MBZ */

struct _wtb_hdr
{
	wt_u8  b_ident[WTB_NIDENT]; /* identification (item indices are WTBI_*) */
	wt_u32 b_size;              /* number of octets to allocate */
	wt_u16 b_shnum;             /* number of sections */
	wt_u16 b_thnum;             /* number of threads */
	wt_u32 b_shoff;             /* absolute offset to section header */
	wt_u32 b_thoff;             /* absolute offset to thread header */
	wt_u32 b_ctor;              /* PC of constructor code */
	wt_u32 b_dtor;              /* PC of destructor code */
};
typedef struct _wtb_hdr wtb_hdr;

#define WTBST_UNUSED           0x00 /* skip this entry */
#define WTBST_CODE             0x01 /* executable program code (at least one; no BOS) */
#define WTBST_DATA             0x02 /* data (if BOS exists, link to WTBST_BOS) */
#define WTBST_BOS              0x03 /* byte order swap (linked to WTBST_DATA; 'sh_addr' is the same) */
#define WTBST_IMPORT           0x05 /* import symbol table */
#define WTBST_EXPORT           0x06 /* export symbol table */
#define WTBST_METADATA         0x07 /* metadata table */
#define WTBST_INTEGRITY        0x08 /* integrity (hash algorithm and digest) */
#define WTBST_DEBUGLABEL       0xB0 /* map addresses to labels */
#define WTBST_CUSTOMLO         0xC0 /* -- */
#define WTBST_CUSTOMHI         0xFF /* -- */

#define WTBSF_CLEAR            0x01 /* if set, clear data on 'sh_addr' */
#define WTBSF_LOAD             0x02 /* if set, load to 'sh_addr' */
#define WTBSF_TYPEVEC          0x04 /* is set, import type vector starts on 'sh_addr' (at most one) */

struct _wtb_shdr
{
	wt_u8  sh_type; /* section type (WTBST_*) */
	wt_u8  sh_flag; /* section flags (WTBSF_*) */
	wt_u16 sh_link; /* link to another section */
	wt_u32 sh_offs; /* absolute offset to data */
	wt_u32 sh_size; /* data size in octets */
	wt_u32 sh_addr; /* relative to program base address */
};
typedef struct _wtb_shdr wtb_shdr;

#define WTB_2CC(_hi_,_lo_)     (((_hi_)<<8) | (_lo_))
#define WTB_4CC(hh,hl,lh,ll)   (((hh)<<24) | ((hl)<<16) | ((lh)<<8) | (ll))

#define WTBTT_USER             WTB_2CC('S', 'R') /* user thread (exactly once, no WTBTF_EXECUTE) */
#define WTBTT_TASK             WTB_2CC('T', 'K') /* user defined task (with WTBTF_EXECUTE) */

#define WTBTF_EXECUTE          0x0001 /* start thread automatically */

struct _wtb_thdr
{
	wt_u16 th_type;   /* thread type (WTBTT_*) */
	wt_u16 th_flag;   /* thread flags (WTBTF_*) */
	wt_u32 th_cssize; /* call stack size */
	wt_u32 th_dssize; /* data stack size */
	wt_u32 th_pcbase; /* start PC */
};
typedef struct _wtb_thdr wtb_thdr;

#define WTBBOS_NOOP            0x00 /* no operation */
#define WTBBOS_NOSWAP_8        0x01 /* no swap, octet count in 8-bit constant */
#define WTBBOS_NOSWAP_16       0x02 /* no swap, octet count in 16-bit constant */
#define WTBBOS_NOSWAP_32       0x03 /* no swap, octet count in 32-bit constant */
#define WTBBOS_NOSWAP_64       0x04 /* no swap, octet count in 64-bit constant */
#define WTBBOS_SWAP16_8        0x11 /* swap 16-bit fields, field count in 8-bit constant */
#define WTBBOS_SWAP16_16       0x12 /* swap 16-bit fields, field count in 16-bit constant */
#define WTBBOS_SWAP16_32       0x13 /* swap 16-bit fields, field count in 32-bit constant */
#define WTBBOS_SWAP16_64       0x14 /* swap 16-bit fields, field count in 64-bit constant */
#define WTBBOS_SWAP32_8        0x21 /* swap 32-bit fields, field count in 8-bit constant */
#define WTBBOS_SWAP32_16       0x22 /* swap 32-bit fields, field count in 16-bit constant */
#define WTBBOS_SWAP32_32       0x23 /* swap 32-bit fields, field count in 32-bit constant */
#define WTBBOS_SWAP32_64       0x24 /* swap 32-bit fields, field count in 64-bit constant */
#define WTBBOS_SWAP64_8        0x31 /* swap 64-bit fields, field count in 8-bit constant */
#define WTBBOS_SWAP64_16       0x32 /* swap 64-bit fields, field count in 16-bit constant */
#define WTBBOS_SWAP64_32       0x33 /* swap 64-bit fields, field count in 32-bit constant */
#define WTBBOS_SWAP64_64       0x34 /* swap 64-bit fields, field count in 64-bit constant */

#define WTBSYMT_FUNCTION       WTB_4CC('f', 'u', 'n', 'c')
#define WTBSYMT_EXCEPTION      WTB_4CC('e', 'x', 'c', ' ')
#define WTBSYMT_TYPE           WTB_4CC('t', 'y', 'p', 'e')
#define WTBSYMT_DATATYPE       WTB_4CC('d', 't', 'y', 'p')
#define WTBSYMT_OPERATOR       WTB_4CC('o', 'p', ' ', ' ')
#define WTBSYMT_RECORD         WTB_4CC('r', 'e', 'c', ' ')
#define WTBSYMT_PRIVATERECORD  WTB_4CC('r', 'e', 'c', 'P')
#define WTBSYMT_LIMITEDRECORD  WTB_4CC('r', 'e', 'c', 'L')
#define WTBSYMT_FUNCTIONTYPE   WTB_4CC('f', 'u', 'n', 'T')
#define WTBSYMT_ENUM           WTB_4CC('e', 'n', 'u', 'm')

struct _wtb_symhdr
{
	wt_u32 symcount;  /* number of wtb_sym */
	wt_u32 reserved;  /* must be zero */
};
typedef struct _wtb_symhdr wtb_symhdr;

#define WTBSYM_OPASSOC(_inf_)  ((_inf_) & 0xff)
#define WTBSYM_OPPREC(_inf_)   (((_inf_)>>8) & 0xff)
#define WTBSYM_CR_OP(_as,_pr)  (((_pr)<<8) | (_as))
#define WTBSYM_FUNCPC(_inf_)   (_inf_)
#define WTBSYM_FUNCOFF(_inf_)  (_inf_)
#define WTBSYM_CR_FUNC(_pc_)   (_pc_)

#define WTBSYM_OP_MINPREC      3
#define WTBSYM_OP_MAXPREC      15
#define WTBSYM_OP_LEFTASSOC    1
#define WTBSYM_OP_RIGHTASSOC   2

/*
	IMPORTANT!

	In case of WTBST_IMPORT and WTBST_EXPORT the 'sh_addr' is
	an offset to the type resolution table of the program.
*/

struct _wtb_sym
{
	wt_u32 sym_type;  /* symbol type (WTBSYMT_*) */
	wt_u32 sym_name;  /* offset to symbol name (relative to wtb_symhdr) */
	wt_u32 sym_offs;  /* offset to symbol data */
	wt_u32 sym_info;  /* additional argument of the type (WTBSYM_*) */
};
typedef struct _wtb_sym wtb_sym;

/*
	Symbol data:

	function:
	functiontype:
		wt_u32 argc;
		wt_u32 argv[];
		wt_u32 retc;
		wt_u32 retv[];
		wt_u8  pure

	record:
		wt_u32 fldc;
		wt_u32 fldv[];

	datatype:
		wt_u32 consnum;
		wt_u32 defcons;
		{
			string cons_name;
			wt_u32 typn;
			wt_u32 typv[];
		} conses [];

	enum:
		wt_u32 enumeratorc;
		string enumeratorv[];

	type:
		wt_u32 type;
*/

/*
	Symbol info:

	function import:
		offset of PFN (relative to program base address)
	function export:
		PC

	operator:
		associativity (WTBSYM_OPASSOC)
		precedence    (WTBSYM_OPPREC)

	record:
		type index

	datatype:
		type index

	enum:
		type index

	functiontype:
		type index

	type:
		type index

	exception import:
		storage address of exception global ID
*/

struct _wtb_metahdr
{
	wt_u32 metacount; /* number of wtb_metahdr */
	wt_u32 reserved;  /* must be zero */
};
typedef struct _wtb_metahdr wtb_metahdr;

struct _wtb_meta
{
	wt_u32 meta_key_offs; /* offset to key name (zero-terminated string) */
	wt_u32 meta_val_offs; /* offset to value (zero-terminated string) */
};
typedef struct _wtb_meta wtb_meta;

struct _wtb_dlabelhdr
{
	wt_u32 labelcount; /* number of wtb_dlabel */
	wt_u32 reserved;   /* must be zero */
};
typedef struct _wtb_dlabelhdr wtb_dlabelhdr;

struct _wtb_dlabel
{
	wt_u32 addr;     /* address in program space */
	wt_u32 lab_offs; /* offset to name (zero-terminated string) */
};
typedef struct _wtb_dlabel wtb_dlabel;

#define WTBHA_BLAKE2_256   0x0101 /* BLAKE2-256 */
#define WTBHA_BLAKE2_384   0x0102 /* BLAKE2-384 */
#define WTBHA_BLAKE2_512   0x0103 /* BLAKE2-512 */
#define WTBHA_SHA2_256     0x0201 /* SHA2-256 */
#define WTBHA_SHA2_384     0x0202 /* SHA2-384 */
#define WTBHA_SHA2_512     0x0203 /* SHA2-512 */

struct _wtb_integhdr
{
	wt_u16 algorithm; /* hash algorithm identifier (WTBHA_*) */
	wt_u16 reserved;  /* must be zero */
};
typedef struct _wtb_integhdr wtb_integhdr;


/*----------------------------------------------------------------*/
/* WT_TOMB_OPCODEs and instruction macros */

#define WT_TOMB_INSTRBITS           16
#define WT_TOMB_OPCODEBITS          8
#define WT_TOMB_OPDATABITS          8
#define WT_TOMB_OPCODEMASK          ((1 << WT_TOMB_OPCODEBITS) - 1)
#define WT_TOMB_OPDATAMASK          ((1 << WT_TOMB_OPDATABITS) - 1)
#define WT_TOMB_HIBITIMM            (1 << (WT_TOMB_OPDATABITS - 1))
#define WT_TOMB_HIBITLONGIMM        (1 << (WT_TOMB_OPDATABITS + WT_TOMB_INSTRBITS - 1))
#define WT_TOMB_MASKLONGIMM         ((1 << (WT_TOMB_OPDATABITS + WT_TOMB_INSTRBITS)) - 1)
#define WT_TOMB_GETOPCODE(_instr_)  ((_instr_) >> (WT_TOMB_INSTRBITS - WT_TOMB_OPDATABITS))
#define WT_TOMB_GETOPDATA(_instr_)  ((_instr_) & WT_TOMB_OPDATAMASK)
#define WT_TOMB_BUILDINSTR(_op,_im) ((((_op) & WT_TOMB_OPCODEMASK) << WT_TOMB_OPDATABITS) | ((_im) & WT_TOMB_OPDATAMASK))

#define WT_TOMB_OPCODE(b0,b1,b2,b3,b4,b5,b6,b7)     (((b0)<<7) | ((b1)<<6) | ((b2)<<5) | ((b3)<<4) | ((b4)<<3) | ((b5)<<2) | ((b6)<<1) | (b7))

enum _wtr_tomb_instructionset
{
	WT_TOMB_OP_NOP      = WT_TOMB_OPCODE(0,0,0,0,0,0,0,0),
/*	                      WT_TOMB_OPCODE(0,0,0,0,0,0,0,1), */
/*	                      WT_TOMB_OPCODE(0,0,0,0,0,0,1,0), */
/*	                      WT_TOMB_OPCODE(0,0,0,0,0,0,1,1), */
/*	                      WT_TOMB_OPCODE(0,0,0,0,0,1,0,0), */
/*	                      WT_TOMB_OPCODE(0,0,0,0,0,1,0,1), */
/*	           LOCK     = WT_TOMB_OPCODE(0,0,0,0,0,1,1,0), */
/*	           ULOCK    = WT_TOMB_OPCODE(0,0,0,0,0,1,1,1), */
	WT_TOMB_OP_LD_I     = WT_TOMB_OPCODE(0,0,0,0,1,0,0,0),
	WT_TOMB_OP_LD_L     = WT_TOMB_OPCODE(0,0,0,0,1,0,0,1),
	WT_TOMB_OP_LD_S     = WT_TOMB_OPCODE(0,0,0,0,1,0,1,0),
/*	                      WT_TOMB_OPCODE(0,0,0,0,1,0,1,1), */
	WT_TOMB_OP_ST_I     = WT_TOMB_OPCODE(0,0,0,0,1,1,0,0),
	WT_TOMB_OP_ST_L     = WT_TOMB_OPCODE(0,0,0,0,1,1,0,1),
	WT_TOMB_OP_ST_S     = WT_TOMB_OPCODE(0,0,0,0,1,1,1,0),
/*	                      WT_TOMB_OPCODE(0,0,0,0,1,1,1,1), */
	WT_TOMB_OP_IDRD_I   = WT_TOMB_OPCODE(0,0,0,1,0,0,0,0),
	WT_TOMB_OP_IDRD_L   = WT_TOMB_OPCODE(0,0,0,1,0,0,0,1),
	WT_TOMB_OP_IDRD_S   = WT_TOMB_OPCODE(0,0,0,1,0,0,1,0),
/*	                      WT_TOMB_OPCODE(0,0,0,1,0,0,1,1), */
	WT_TOMB_OP_PBA_I    = WT_TOMB_OPCODE(0,0,0,1,0,1,0,0),
	WT_TOMB_OP_PBA_L    = WT_TOMB_OPCODE(0,0,0,1,0,1,0,1),
	WT_TOMB_OP_PBA_S    = WT_TOMB_OPCODE(0,0,0,1,0,1,1,0),
/*	                      WT_TOMB_OPCODE(0,0,0,1,0,1,1,1), */
	WT_TOMB_OP_PVT_I    = WT_TOMB_OPCODE(0,0,0,1,1,0,0,0),
	WT_TOMB_OP_PVT_L    = WT_TOMB_OPCODE(0,0,0,1,1,0,0,1),
	WT_TOMB_OP_PVT_S    = WT_TOMB_OPCODE(0,0,0,1,1,0,1,0),
	WT_TOMB_OP_PPC      = WT_TOMB_OPCODE(0,0,0,1,1,0,1,1),
	WT_TOMB_OP_DUP_I    = WT_TOMB_OPCODE(0,0,0,1,1,1,0,0),
	WT_TOMB_OP_DUP_L    = WT_TOMB_OPCODE(0,0,0,1,1,1,0,1),
	WT_TOMB_OP_POP_I    = WT_TOMB_OPCODE(0,0,0,1,1,1,1,0),
	WT_TOMB_OP_POP_L    = WT_TOMB_OPCODE(0,0,0,1,1,1,1,1),
	WT_TOMB_OP_PWZ_I    = WT_TOMB_OPCODE(0,0,1,0,0,0,0,0),
	WT_TOMB_OP_PWZ_L    = WT_TOMB_OPCODE(0,0,1,0,0,0,0,1),
	WT_TOMB_OP_PWO_I    = WT_TOMB_OPCODE(0,0,1,0,0,0,1,0),
	WT_TOMB_OP_PWO_L    = WT_TOMB_OPCODE(0,0,1,0,0,0,1,1),
	WT_TOMB_OP_PWZL_I   = WT_TOMB_OPCODE(0,0,1,0,0,1,0,0),
	WT_TOMB_OP_PWZL_L   = WT_TOMB_OPCODE(0,0,1,0,0,1,0,1),
	WT_TOMB_OP_PWOL_I   = WT_TOMB_OPCODE(0,0,1,0,0,1,1,0),
	WT_TOMB_OP_PWOL_L   = WT_TOMB_OPCODE(0,0,1,0,0,1,1,1),
	WT_TOMB_OP_DVTA_I   = WT_TOMB_OPCODE(0,0,1,0,1,0,0,0),
	WT_TOMB_OP_DVTA_L   = WT_TOMB_OPCODE(0,0,1,0,1,0,0,1),
	WT_TOMB_OP_DVTA_S   = WT_TOMB_OPCODE(0,0,1,0,1,0,1,0),
	WT_TOMB_OP_DVTF     = WT_TOMB_OPCODE(0,0,1,0,1,0,1,1),
	WT_TOMB_OP_OSA_I    = WT_TOMB_OPCODE(0,0,1,0,1,1,0,0),
	WT_TOMB_OP_OSA_L    = WT_TOMB_OPCODE(0,0,1,0,1,1,0,1),
	WT_TOMB_OP_OSA_S    = WT_TOMB_OPCODE(0,0,1,0,1,1,1,0),
	WT_TOMB_OP_OSF      = WT_TOMB_OPCODE(0,0,1,0,1,1,1,1),
	WT_TOMB_OP_OSPU     = WT_TOMB_OPCODE(0,0,1,1,0,0,0,0),
	WT_TOMB_OP_OSPO     = WT_TOMB_OPCODE(0,0,1,1,0,0,0,1),
	WT_TOMB_OP_OSSP     = WT_TOMB_OPCODE(0,0,1,1,0,0,1,0),
	WT_TOMB_OP_TSZ      = WT_TOMB_OPCODE(0,0,1,1,0,0,1,1),
	WT_TOMB_OP_TPU      = WT_TOMB_OPCODE(0,0,1,1,0,1,0,0),
	WT_TOMB_OP_TPO      = WT_TOMB_OPCODE(0,0,1,1,0,1,0,1),
	WT_TOMB_OP_TDR      = WT_TOMB_OPCODE(0,0,1,1,0,1,1,0),
	WT_TOMB_OP_TCL      = WT_TOMB_OPCODE(0,0,1,1,0,1,1,1),
	WT_TOMB_OP_PDSP_I   = WT_TOMB_OPCODE(0,0,1,1,1,0,0,0),
	WT_TOMB_OP_PDSP_L   = WT_TOMB_OPCODE(0,0,1,1,1,0,0,1),
	WT_TOMB_OP_VAS_I    = WT_TOMB_OPCODE(0,0,1,1,1,0,1,0),
	WT_TOMB_OP_VAS_L    = WT_TOMB_OPCODE(0,0,1,1,1,0,1,1),
	WT_TOMB_OP_ENTER_I  = WT_TOMB_OPCODE(0,0,1,1,1,1,0,0),
	WT_TOMB_OP_ENTER_L  = WT_TOMB_OPCODE(0,0,1,1,1,1,0,1),
	WT_TOMB_OP_ENTER_S  = WT_TOMB_OPCODE(0,0,1,1,1,1,1,0),
	WT_TOMB_OP_LEAVE    = WT_TOMB_OPCODE(0,0,1,1,1,1,1,1),
	WT_TOMB_OP_JMP_I    = WT_TOMB_OPCODE(0,1,0,0,0,0,0,0),
	WT_TOMB_OP_JMP_L    = WT_TOMB_OPCODE(0,1,0,0,0,0,0,1),
	WT_TOMB_OP_JMP_S    = WT_TOMB_OPCODE(0,1,0,0,0,0,1,0),
/*	                      WT_TOMB_OPCODE(0,1,0,0,0,0,1,1), */
	WT_TOMB_OP_JT_I     = WT_TOMB_OPCODE(0,1,0,0,0,1,0,0),
	WT_TOMB_OP_JT_L     = WT_TOMB_OPCODE(0,1,0,0,0,1,0,1),
	WT_TOMB_OP_JT_S     = WT_TOMB_OPCODE(0,1,0,0,0,1,1,0),
/*	                      WT_TOMB_OPCODE(0,1,0,0,0,1,1,1), */
	WT_TOMB_OP_JF_I     = WT_TOMB_OPCODE(0,1,0,0,1,0,0,0),
	WT_TOMB_OP_JF_L     = WT_TOMB_OPCODE(0,1,0,0,1,0,0,1),
	WT_TOMB_OP_JF_S     = WT_TOMB_OPCODE(0,1,0,0,1,0,1,0),
/*	                      WT_TOMB_OPCODE(0,1,0,0,1,0,1,1), */
	WT_TOMB_OP_JA_I     = WT_TOMB_OPCODE(0,1,0,0,1,1,0,0),
	WT_TOMB_OP_JA_L     = WT_TOMB_OPCODE(0,1,0,0,1,1,0,1),
	WT_TOMB_OP_JA_S     = WT_TOMB_OPCODE(0,1,0,0,1,1,1,0),
/*	                      WT_TOMB_OPCODE(0,1,0,0,1,1,1,1), */
	WT_TOMB_OP_EXEC_I   = WT_TOMB_OPCODE(0,1,0,1,0,0,0,0),
	WT_TOMB_OP_EXEC_L   = WT_TOMB_OPCODE(0,1,0,1,0,0,0,1),
	WT_TOMB_OP_EXEC_S   = WT_TOMB_OPCODE(0,1,0,1,0,0,1,0),
/*	                      WT_TOMB_OPCODE(0,1,0,1,0,0,1,1), */
	WT_TOMB_OP_DLE_I    = WT_TOMB_OPCODE(0,1,0,1,0,1,0,0),
	WT_TOMB_OP_DLE_L    = WT_TOMB_OPCODE(0,1,0,1,0,1,0,1),
	WT_TOMB_OP_DLE_S    = WT_TOMB_OPCODE(0,1,0,1,0,1,1,0),
	WT_TOMB_OP_CFP      = WT_TOMB_OPCODE(0,1,0,1,0,1,1,1),
	WT_TOMB_OP_MKFP_I   = WT_TOMB_OPCODE(0,1,0,1,1,0,0,0),
	WT_TOMB_OP_MKFP_L   = WT_TOMB_OPCODE(0,1,0,1,1,0,0,1),
	WT_TOMB_OP_MKFP_S   = WT_TOMB_OPCODE(0,1,0,1,1,0,1,0),
	WT_TOMB_OP_TTD      = WT_TOMB_OPCODE(0,1,0,1,1,0,1,1), /* since v3.0.0.2 */
	WT_TOMB_OP_LDFP     = WT_TOMB_OPCODE(0,1,0,1,1,1,0,0),
	WT_TOMB_OP_STFP     = WT_TOMB_OPCODE(0,1,0,1,1,1,0,1),
	WT_TOMB_OP_FPS_I    = WT_TOMB_OPCODE(0,1,0,1,1,1,1,0),
	WT_TOMB_OP_FPS_L    = WT_TOMB_OPCODE(0,1,0,1,1,1,1,1),
	WT_TOMB_OP_STRN     = WT_TOMB_OPCODE(0,1,1,0,0,0,0,0),
	WT_TOMB_OP_STRC     = WT_TOMB_OPCODE(0,1,1,0,0,0,0,1),
	WT_TOMB_OP_STRL     = WT_TOMB_OPCODE(0,1,1,0,0,0,1,0),
	WT_TOMB_OP_STRD     = WT_TOMB_OPCODE(0,1,1,0,0,0,1,1),
	WT_TOMB_OP_STRCR_I  = WT_TOMB_OPCODE(0,1,1,0,0,1,0,0),
	WT_TOMB_OP_STRCR_L  = WT_TOMB_OPCODE(0,1,1,0,0,1,0,1),
	WT_TOMB_OP_STRCR_S  = WT_TOMB_OPCODE(0,1,1,0,0,1,1,0),
	WT_TOMB_OP_SEQN_I   = WT_TOMB_OPCODE(0,1,1,0,0,1,1,1),
	WT_TOMB_OP_SEQN_L   = WT_TOMB_OPCODE(0,1,1,0,1,0,0,0),
	WT_TOMB_OP_SEQN_S   = WT_TOMB_OPCODE(0,1,1,0,1,0,0,1),
	WT_TOMB_OP_SEQD     = WT_TOMB_OPCODE(0,1,1,0,1,0,1,0),
	WT_TOMB_OP_SEQIR_I  = WT_TOMB_OPCODE(0,1,1,0,1,0,1,1),
	WT_TOMB_OP_SEQIR_L  = WT_TOMB_OPCODE(0,1,1,0,1,1,0,0),
	WT_TOMB_OP_SEQIR_S  = WT_TOMB_OPCODE(0,1,1,0,1,1,0,1),
	WT_TOMB_OP_SEQL     = WT_TOMB_OPCODE(0,1,1,0,1,1,1,0),
	WT_TOMB_OP_RECN_I   = WT_TOMB_OPCODE(0,1,1,0,1,1,1,1),
	WT_TOMB_OP_RECN_L   = WT_TOMB_OPCODE(0,1,1,1,0,0,0,0),
	WT_TOMB_OP_RECN_S   = WT_TOMB_OPCODE(0,1,1,1,0,0,0,1),
	WT_TOMB_OP_RECD     = WT_TOMB_OPCODE(0,1,1,1,0,0,1,0),
	WT_TOMB_OP_RECFR_I  = WT_TOMB_OPCODE(0,1,1,1,0,0,1,1),
	WT_TOMB_OP_RECFR_L  = WT_TOMB_OPCODE(0,1,1,1,0,1,0,0),
	WT_TOMB_OP_RECFR_S  = WT_TOMB_OPCODE(0,1,1,1,0,1,0,1),
/*	                      WT_TOMB_OPCODE(0,1,1,1,0,1,1,0), */
/*	                      WT_TOMB_OPCODE(0,1,1,1,0,1,1,1), */
/*	                      WT_TOMB_OPCODE(0,1,1,1,1,0,0,0), */
/*	                      WT_TOMB_OPCODE(0,1,1,1,1,0,0,1), */
/*	                      WT_TOMB_OPCODE(0,1,1,1,1,0,1,0), */
	WT_TOMB_OP_ERAISE   = WT_TOMB_OPCODE(0,1,1,1,1,0,1,1),
	WT_TOMB_OP_EXHAN_I  = WT_TOMB_OPCODE(0,1,1,1,1,1,0,0),
	WT_TOMB_OP_EXHAN_L  = WT_TOMB_OPCODE(0,1,1,1,1,1,0,1),
	WT_TOMB_OP_EXHAN_S  = WT_TOMB_OPCODE(0,1,1,1,1,1,1,0),
	WT_TOMB_OP_EHCL     = WT_TOMB_OPCODE(0,1,1,1,1,1,1,1),
	WT_TOMB_OP_ADD_I    = WT_TOMB_OPCODE(1,0,0,0,0,0,0,0),
	WT_TOMB_OP_ADD_S    = WT_TOMB_OPCODE(1,0,0,0,0,0,0,1),
	WT_TOMB_OP_SUB_I    = WT_TOMB_OPCODE(1,0,0,0,0,0,1,0),
	WT_TOMB_OP_SUB_S    = WT_TOMB_OPCODE(1,0,0,0,0,0,1,1),
	WT_TOMB_OP_MUL_I    = WT_TOMB_OPCODE(1,0,0,0,0,1,0,0),
	WT_TOMB_OP_MUL_S    = WT_TOMB_OPCODE(1,0,0,0,0,1,0,1),
	WT_TOMB_OP_DIV_I    = WT_TOMB_OPCODE(1,0,0,0,0,1,1,0),
	WT_TOMB_OP_DIV_S    = WT_TOMB_OPCODE(1,0,0,0,0,1,1,1),
	WT_TOMB_OP_SHL_I    = WT_TOMB_OPCODE(1,0,0,0,1,0,0,0),
	WT_TOMB_OP_SHL_S    = WT_TOMB_OPCODE(1,0,0,0,1,0,0,1),
	WT_TOMB_OP_SHR_I    = WT_TOMB_OPCODE(1,0,0,0,1,0,1,0),
	WT_TOMB_OP_SHR_S    = WT_TOMB_OPCODE(1,0,0,0,1,0,1,1),
	WT_TOMB_OP_SAR_I    = WT_TOMB_OPCODE(1,0,0,0,1,1,0,0),
	WT_TOMB_OP_SAR_S    = WT_TOMB_OPCODE(1,0,0,0,1,1,0,1),
	WT_TOMB_OP_ROR_I    = WT_TOMB_OPCODE(1,0,0,0,1,1,1,0),
	WT_TOMB_OP_ROR_S    = WT_TOMB_OPCODE(1,0,0,0,1,1,1,1),
	WT_TOMB_OP_SMUL_I   = WT_TOMB_OPCODE(1,0,0,1,0,0,0,0),
	WT_TOMB_OP_SMUL_S   = WT_TOMB_OPCODE(1,0,0,1,0,0,0,1),
	WT_TOMB_OP_SDIV_I   = WT_TOMB_OPCODE(1,0,0,1,0,0,1,0),
	WT_TOMB_OP_SDIV_S   = WT_TOMB_OPCODE(1,0,0,1,0,0,1,1),
	WT_TOMB_OP_MOD      = WT_TOMB_OPCODE(1,0,0,1,0,1,0,0),
	WT_TOMB_OP_SMOD     = WT_TOMB_OPCODE(1,0,0,1,0,1,0,1),
	WT_TOMB_OP_LSOR_I   = WT_TOMB_OPCODE(1,0,0,1,0,1,1,0),
	WT_TOMB_OP_LSOR_L   = WT_TOMB_OPCODE(1,0,0,1,0,1,1,1),
	WT_TOMB_OP_RSB_I    = WT_TOMB_OPCODE(1,0,0,1,1,0,0,0),
	WT_TOMB_OP_RSB_L    = WT_TOMB_OPCODE(1,0,0,1,1,0,0,1),
	WT_TOMB_OP_ADDL     = WT_TOMB_OPCODE(1,0,0,1,1,0,1,0),
	WT_TOMB_OP_SUBL     = WT_TOMB_OPCODE(1,0,0,1,1,0,1,1),
	WT_TOMB_OP_MULL     = WT_TOMB_OPCODE(1,0,0,1,1,1,0,0),
	WT_TOMB_OP_DIVL     = WT_TOMB_OPCODE(1,0,0,1,1,1,0,1),
	WT_TOMB_OP_SHLL     = WT_TOMB_OPCODE(1,0,0,1,1,1,1,0),
	WT_TOMB_OP_SHRL     = WT_TOMB_OPCODE(1,0,0,1,1,1,1,1),
	WT_TOMB_OP_SARL     = WT_TOMB_OPCODE(1,0,1,0,0,0,0,0),
	WT_TOMB_OP_RORL     = WT_TOMB_OPCODE(1,0,1,0,0,0,0,1),
	WT_TOMB_OP_XOR      = WT_TOMB_OPCODE(1,0,1,0,0,0,1,0),
	WT_TOMB_OP_BAND     = WT_TOMB_OPCODE(1,0,1,0,0,0,1,1),
	WT_TOMB_OP_BOR      = WT_TOMB_OPCODE(1,0,1,0,0,1,0,0),
	WT_TOMB_OP_BNOT     = WT_TOMB_OPCODE(1,0,1,0,0,1,0,1),
	WT_TOMB_OP_XORL     = WT_TOMB_OPCODE(1,0,1,0,0,1,1,0),
	WT_TOMB_OP_BANDL    = WT_TOMB_OPCODE(1,0,1,0,0,1,1,1),
	WT_TOMB_OP_BORL     = WT_TOMB_OPCODE(1,0,1,0,1,0,0,0),
	WT_TOMB_OP_BNOTL    = WT_TOMB_OPCODE(1,0,1,0,1,0,0,1),
	WT_TOMB_OP_LSORL_I  = WT_TOMB_OPCODE(1,0,1,0,1,0,1,0),
	WT_TOMB_OP_LSORL_L  = WT_TOMB_OPCODE(1,0,1,0,1,0,1,1),
	WT_TOMB_OP_SMULL    = WT_TOMB_OPCODE(1,0,1,0,1,1,0,0),
	WT_TOMB_OP_SDIVL    = WT_TOMB_OPCODE(1,0,1,0,1,1,0,1),
	WT_TOMB_OP_MODL     = WT_TOMB_OPCODE(1,0,1,0,1,1,1,0),
	WT_TOMB_OP_SMODL    = WT_TOMB_OPCODE(1,0,1,0,1,1,1,1),
	WT_TOMB_OP_NEG      = WT_TOMB_OPCODE(1,0,1,1,0,0,0,0),
	WT_TOMB_OP_NEGL     = WT_TOMB_OPCODE(1,0,1,1,0,0,0,1),
	WT_TOMB_OP_FADD     = WT_TOMB_OPCODE(1,0,1,1,0,0,1,0),
	WT_TOMB_OP_FSUB     = WT_TOMB_OPCODE(1,0,1,1,0,0,1,1),
	WT_TOMB_OP_FMUL     = WT_TOMB_OPCODE(1,0,1,1,0,1,0,0),
	WT_TOMB_OP_FDIV     = WT_TOMB_OPCODE(1,0,1,1,0,1,0,1),
	WT_TOMB_OP_FADDL    = WT_TOMB_OPCODE(1,0,1,1,0,1,1,0),
	WT_TOMB_OP_FSUBL    = WT_TOMB_OPCODE(1,0,1,1,0,1,1,1),
	WT_TOMB_OP_FMULL    = WT_TOMB_OPCODE(1,0,1,1,1,0,0,0),
	WT_TOMB_OP_FDIVL    = WT_TOMB_OPCODE(1,0,1,1,1,0,0,1),
	WT_TOMB_OP_FNEG     = WT_TOMB_OPCODE(1,0,1,1,1,0,1,0),
	WT_TOMB_OP_FNEGL    = WT_TOMB_OPCODE(1,0,1,1,1,0,1,1),
	WT_TOMB_OP_FMOD     = WT_TOMB_OPCODE(1,0,1,1,1,1,0,0),
	WT_TOMB_OP_FMODL    = WT_TOMB_OPCODE(1,0,1,1,1,1,0,1),
	WT_TOMB_OP_FABS     = WT_TOMB_OPCODE(1,0,1,1,1,1,1,0),
	WT_TOMB_OP_FABSL    = WT_TOMB_OPCODE(1,0,1,1,1,1,1,1),
	WT_TOMB_OP_CMP      = WT_TOMB_OPCODE(1,1,0,0,0,0,0,0), /* result: { -1, 0, 1 } */
	WT_TOMB_OP_SCMP     = WT_TOMB_OPCODE(1,1,0,0,0,0,0,1),
	WT_TOMB_OP_CMPL     = WT_TOMB_OPCODE(1,1,0,0,0,0,1,0),
	WT_TOMB_OP_SCMPL    = WT_TOMB_OPCODE(1,1,0,0,0,0,1,1),
	WT_TOMB_OP_FCMP     = WT_TOMB_OPCODE(1,1,0,0,0,1,0,0),
	WT_TOMB_OP_FCMPL    = WT_TOMB_OPCODE(1,1,0,0,0,1,0,1),
	WT_TOMB_OP_SCMPB    = WT_TOMB_OPCODE(1,1,0,0,0,1,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SCMPH    = WT_TOMB_OPCODE(1,1,0,0,0,1,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_QLT      = WT_TOMB_OPCODE(1,1,0,0,1,0,0,0), /* {-1, 0, 1} -> {0, 1} */
	WT_TOMB_OP_QLE      = WT_TOMB_OPCODE(1,1,0,0,1,0,0,1),
	WT_TOMB_OP_QGT      = WT_TOMB_OPCODE(1,1,0,0,1,0,1,0),
	WT_TOMB_OP_QGE      = WT_TOMB_OPCODE(1,1,0,0,1,0,1,1),
	WT_TOMB_OP_QEQ      = WT_TOMB_OPCODE(1,1,0,0,1,1,0,0),
	WT_TOMB_OP_QNE      = WT_TOMB_OPCODE(1,1,0,0,1,1,0,1),
	WT_TOMB_OP_OSDR     = WT_TOMB_OPCODE(1,1,0,0,1,1,1,0), /* since v3.0.0.1 */
	WT_TOMB_OP_MSKBTM   = WT_TOMB_OPCODE(1,1,0,0,1,1,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_LNOT     = WT_TOMB_OPCODE(1,1,0,1,0,0,0,0),
	WT_TOMB_OP_LAND     = WT_TOMB_OPCODE(1,1,0,1,0,0,0,1),
	WT_TOMB_OP_LOR      = WT_TOMB_OPCODE(1,1,0,1,0,0,1,0),
	WT_TOMB_OP_LXOR     = WT_TOMB_OPCODE(1,1,0,1,0,0,1,1),
	WT_TOMB_OP_LDB      = WT_TOMB_OPCODE(1,1,0,1,0,1,0,0),
	WT_TOMB_OP_STB      = WT_TOMB_OPCODE(1,1,0,1,0,1,0,1),
	WT_TOMB_OP_LDH      = WT_TOMB_OPCODE(1,1,0,1,0,1,1,0),
	WT_TOMB_OP_STH      = WT_TOMB_OPCODE(1,1,0,1,0,1,1,1),
	WT_TOMB_OP_ADDB_I   = WT_TOMB_OPCODE(1,1,0,1,1,0,0,0), /* since v3.0.0.3 */
	WT_TOMB_OP_ADDB_S   = WT_TOMB_OPCODE(1,1,0,1,1,0,0,1), /* since v3.0.0.3 */
	WT_TOMB_OP_SUBB_I   = WT_TOMB_OPCODE(1,1,0,1,1,0,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SUBB_S   = WT_TOMB_OPCODE(1,1,0,1,1,0,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_MULB     = WT_TOMB_OPCODE(1,1,0,1,1,1,0,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SMULB    = WT_TOMB_OPCODE(1,1,0,1,1,1,0,1), /* since v3.0.0.3 */
	WT_TOMB_OP_DIVB     = WT_TOMB_OPCODE(1,1,0,1,1,1,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SDIVB    = WT_TOMB_OPCODE(1,1,0,1,1,1,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_MODB     = WT_TOMB_OPCODE(1,1,1,0,0,0,0,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SMODB    = WT_TOMB_OPCODE(1,1,1,0,0,0,0,1), /* since v3.0.0.3 */
	WT_TOMB_OP_SHLB_I   = WT_TOMB_OPCODE(1,1,1,0,0,0,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SHLB_S   = WT_TOMB_OPCODE(1,1,1,0,0,0,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_RORB     = WT_TOMB_OPCODE(1,1,1,0,0,1,0,0), /* since v3.0.0.3 */
	WT_TOMB_OP_BNOTB    = WT_TOMB_OPCODE(1,1,1,0,0,1,0,1), /* since v3.0.0.3 */
	WT_TOMB_OP_ADDH_I   = WT_TOMB_OPCODE(1,1,1,0,0,1,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_ADDH_S   = WT_TOMB_OPCODE(1,1,1,0,0,1,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_SUBH_I   = WT_TOMB_OPCODE(1,1,1,0,1,0,0,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SUBH_S   = WT_TOMB_OPCODE(1,1,1,0,1,0,0,1), /* since v3.0.0.3 */
	WT_TOMB_OP_MULH     = WT_TOMB_OPCODE(1,1,1,0,1,0,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SMULH    = WT_TOMB_OPCODE(1,1,1,0,1,0,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_DIVH     = WT_TOMB_OPCODE(1,1,1,0,1,1,0,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SDIVH    = WT_TOMB_OPCODE(1,1,1,0,1,1,0,1), /* since v3.0.0.3 */
	WT_TOMB_OP_MODH     = WT_TOMB_OPCODE(1,1,1,0,1,1,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SMODH    = WT_TOMB_OPCODE(1,1,1,0,1,1,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_SHLH_I   = WT_TOMB_OPCODE(1,1,1,1,0,0,0,0), /* since v3.0.0.3 */
	WT_TOMB_OP_SHLH_S   = WT_TOMB_OPCODE(1,1,1,1,0,0,0,1), /* since v3.0.0.3 */
	WT_TOMB_OP_RORH     = WT_TOMB_OPCODE(1,1,1,1,0,0,1,0), /* since v3.0.0.3 */
	WT_TOMB_OP_BNOTH    = WT_TOMB_OPCODE(1,1,1,1,0,0,1,1), /* since v3.0.0.3 */
	WT_TOMB_OP_SGNXD    = WT_TOMB_OPCODE(1,1,1,1,0,1,0,0), /* since v3.0.0.3 */
/*	                      WT_TOMB_OPCODE(1,1,1,1,0,1,0,1), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,0,1,1,0), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,0,1,1,1), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,0,0,0), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,0,0,1), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,0,1,0), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,0,1,1), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,1,0,0), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,1,0,1), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,1,1,0), */
/*	                      WT_TOMB_OPCODE(1,1,1,1,1,1,1,1), */
};
typedef enum _wtr_tomb_instructionset wtr_tomb_instructionset;


/*----------------------------------------------------------------*/
/* runtime types */

#ifndef WT_DISABLE_PRAGMA_PACK
/* TODO */
#  pragma pack(1)
#endif /* WT_DISABLE_PRAGMA_PACK */
struct _wt_pfn
{
	union
	{
	void*  ptr;
	wt_u64 _pad;
	};
	wt_u32 ext;
};
typedef struct _wt_pfn wt_pfn;
#ifndef WT_DISABLE_PRAGMA_PACK
/* TODO */
#  pragma pack()
#endif /* WT_DISABLE_PRAGMA_PACK */

struct wtr_profile;
struct wtr_program;
struct wtr_function;
struct wtr_context;
struct wtr_extcallinfo;

typedef struct wtr_profile*  wt_profile;
typedef struct wtr_program*  wt_program;
typedef struct wtr_function* wt_function;


/*----------------------------------------------------------------*/
/* declarations for the VM */

typedef void* wt_file;
typedef void* wt_dir;
typedef void* wt_socket;

struct _wt_value
{
	union
	{
	const wt_u32 type;
	wt_u32       writable_type;
	void*        _pad;
	};
	union
	{
	wt_bool      v_bool;
	wt_string    v_string;
	wt_i8        v_i8;
	wt_u8        v_u8;
	wt_i16       v_i16;
	wt_u16       v_u16;
	wt_i32       v_i32;
	wt_u32       v_u32;
	wt_i64       v_i64;
	wt_u64       v_u64;
#if WTRUNTIME_ENABLE_FLOAT32
	wt_f32       v_f32;
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
	wt_f64       v_f64;
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
	wt_char      v_char;
	wt_handle    v_handle;
	wt_file      v_file;
	wt_dir       v_dir;
	wt_socket    v_socket;
	struct
	{
	wt_u32 n;
	void*  p;
	}            v_seq;
	};
};
typedef struct _wt_value wt_value;

/* External function. Implementation shall call exactly one of the */
/* following functions to complete the call: */
/* - wtrContextPushReturnValues(), */
/* - wtrContextRaiseException(); */
/* or: not calling any of the function above to yield from the external */
/* function and cause WTRR_EXTCALL to be returned from wtrContextRun() */
/* until the call is completed. */
typedef void (*WT_PFN_EXTFUNC)(struct wtr_context* ctx,
	struct wtr_extcallinfo* callinfo, wt_value* argv);

/* realloc()-style function with context. If 'memory' is NULL, then should */
/* allocate a memory block with the specified size; if 'size' is zero, then */
/* should deallocate the memory block; if both zero (NULL and zero), then */
/* nothing should be changed; if 'memory' is not NULL and 'size' is not */
/* zero, then realloc can happen (can return NULL if not supported). */
typedef void* (*WT_PFN_ALLOC)(void* ctx, void* memory, size_t size);

struct wtr_loaderops
{
	void (*check_failed)(void* ctx, unsigned int line);
	void (*unres_exc)(void* ctx, const char* name);
	void (*unres_func)(void* ctx, const char* name, wt_u32 argc, wt_u32 retc,
		const wt_u32* argv, const wt_u32* retv);
	void (*unres_type)(void* ctx, const char* name);
	void (*unres_op)(void* ctx, const char* name);
};

enum wtr_runresult
{
	/* initial state, the context is not yet set up to execute anything; */
	/* use wtrContextInvokeProgram() or wtrContextInvokeFunction() */
	WTRR_INIT = 0,

	/* there are instructions to be executed */
	WTRR_RUN,

	/* called function returned, pop return values with */
	/* wtrContextPopReturnValues(), even if number of returned values is */
	/* zero -- this will reset the context to initial state */
	WTRR_RETURNED,

	/* function consumed supplied quantum, subsequent run possible */
	WTRR_QUANTUM,

	/* internal state, just call wtrContextRun() -- the next action is to */
	/* perform to raise the scheduled exception */
	WTRR_RAISE,

	/* called function returned with an unhandled exception -- take the */
	/* exception ID with wtrContextGetException() */
	WTRR_EXCEPTION,

	/* execution suspended due to a standard I/O call -- gather info with */
	/* wtrContextGetIOCallParams(), and push return values using */
	/* wtrContextPushReturnValues() even if there are no actual return */
	/* values; or raise an exception with wtrContextRaiseException() */
	WTRR_IOCALL,

	/* execution suspended due to an external function call -- gather info */
	/* with wtrContextGetExtCallParams(), and push return values using */
	/* wtrContextPushReturnValues() even if there are no actual return */
	/* values; raise an exception with wtrContextRaiseException(), or */
	/* yield by not calling any of the function mentioned above [in this */
	/* case wtrContextRun() will return, and the external call can be */
	/* handled at the callers side] */
	WTRR_EXTCALL,
};

/* Number of arguments the I/O call with the most arguments has. */
#define WTR_IOCALL_ARGMAX       4

enum wtr_iocall
{
	/* write to standard output */
	/* (string buf, u32 len) -> () */
	/* Host program shall write the specified string to the standard output */
	/* by the specified length. Should not be treated as a C-string. */
	WTIOC_WRITESTDOUT,

	/* read a line from standard input */
	/* (string buf, u32 maxlen, char separator) -> (u32 len) */
	/* Host program shall read a line from the standard input, which must */
	/* meet the following conditions: number of read characters must not be */
	/* more than 'maxlen' and read shall stop when 'separator' character is */
	/* read (and this character must not be part of the line). If line does */
	/* not fit in the supplied buffer 'IOException' shall be raised. If the */
	/* input is empty (and no characters were read), host must raise an */
	/* 'EndOfFileException'. Return value is number of characters filled */
	/* into the buffer. */
	WTIOC_READLNSTDIN,

	/* check permissions for a file */
	/* (string pathname, u32 mode) -> (bool) */
	/* Mode is OR'd combination of WTFA_* constants. */
	WTIOC_ACCESS,

	/* get file status */
	/* (string) -> (bool, char, u64, u64, u64) */
	/* Parameter is the path name of the file. Return values are: */
	/* - bool result: true if status successfully acquired, false otherwise; */
	/* - char type: type of the file, WTDT_... constants; */
	/* - u64 size: size of the file; */
	/* - u64 mtime: time of last modification (in msecs since epoch); */
	/* - u64 birthtime: time of birth (in msecs since epoch). */
	WTIOC_STAT,

	/* get file status */
	/* (file) -> (bool, char, u64, u64, u64) */
	/* Same as WTIOC_STAT, but works on open files. */
	WTIOC_FSTAT,

	/* open a file */
	/* (string pathname, char mode) -> (file) */
	/* Mode is one of 'r', 'R', 'w', 'W', 'a', 'A'. The host program shall */
	/* raise IOException if the file cannot be opened. */
	WTIOC_FOPEN,

	/* close a file */
	/* (file) -> () */
	/* Close the file, if it was opened. */
	WTIOC_FCLOSE,

	/* reposition a stream */
	/* (file, i64 offset, u32 whence) -> (i64 result) */
	/* Value in 'whence' is one of the WTSW_... macros (in case of an */
	/* invalid value, IOException should be raised). Seeking in the file */
	/* should clear end-of-file indicator. Return value should be the */
	/* resultant file offset. */
	WTIOC_FSEEK,

	/* check end-of-file indicator */
	/* (file) -> (bool) */
	/* Host should return true if EOF is reached, false otherwise. Seeking */
	/* in the file should clear end-of-file indicator. */
	WTIOC_FEOF,

	/* write to file */
	/* (file, string buf, u32 len) -> (u32 len) */
	/* Write the specified buffer into the file. Buffer shall not be */
	/* treated as a C-string.  */
	WTIOC_FWRITE,

	/* read from file */
	/* (file, string buf, u32 len) -> (u32 len) */
	/* Read the specified amount of data into the buffer. Return value is */
	/* the actually read number of bytes; short-read is allowed. */
	WTIOC_FREAD,

	/* read a line from file */
	/* (file, string buf, u32 maxlen, char separator) -> (u32 len) */
	/* Same as WTIOC_READLNSTDIN, but works on opened files. */
	WTIOC_FREADLN,

	/* open a directory */
	/* (string pathname) -> (dir) */
	/* The host program shall raise 'IOException' if the directory cannot */
	/* be opened. */
	WTIOC_OPENDIR,

	/* close a directory */
	/* (dir) -> () */
	/* Close the directory, if it was opened. (If it wasn't opened, then */
	/* no operation.) */
	WTIOC_CLOSEDIR,

	/* read a directory */
	/* (dir) -> (bool) */
	/* Read the next entry in the directory. Return value shall be true if */
	/* there was a read entry; false otherwise. */
	WTIOC_READDIR,

	/* get type of the current directory entry */
	/* (dir) -> (char) */
	/* Return value must be a WTDT_... constant which indicates the type */
	/* of the current directory entry. Raising IOException is allowed if no */
	/* entry was read. */
	WTIOC_DIRGETTYPE,

	/* get the current directory entry name */
	/* (dir, string buf, u32 maxlen) -> (u32) */
	/* Fill the name of the current directory entry into 'buf'. If no entry */
	/* was previously read, or name is longer than 'maxlen', an IOException */
	/* shall be raised. Return the name length. */
	WTIOC_DIRGETNAME,
};

struct wtr_extcallinfo
{
	wt_profile     prof;  /* the virtual machine */
	WT_PFN_EXTFUNC func;  /* pointer to the external function */
	void*          token; /* user supplied function token */
	wt_u32         argc;  /* number of arguments */
	wt_u32         retc;  /* number of return values */
	const wt_u32*  retv;  /* return value type array */
};


/*----------------------------------------------------------------*/

// TODO: remove
typedef wt_error (*WT_PFN_PROFILEGETFUNCTION)(wt_profile profile, wt_function* p_func, const char* name, wt_u32 argc, wt_u32* argv, wt_u32 retc, wt_u32* retv, wt_bool pure);
typedef wt_error (*WT_PFN_PROFILERELEASEFUNCTION)(wt_profile profile, wt_function func);

/*----------------------------------------------------------------*/
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */
/*----------------------------------------------------------------*/


/* Get the Welltype runtime version. Version number is a hexadecimal */
/* number made with WT_MAKEVERSION(). Returned version is comparable */
/* to WTR_VERSION. */
/* Returns: version number */
WTAPIENTRY wt_u32 wtrGetVersion(void);


/* Convert error code to its mnemonic (i.e. WTE_OK -> "OK"). */
/* Returns: string or NULL */
/*   err - error code */
WTAPIENTRY const char* wtrMnError(wt_error err);


/* Convert error code to string (i.e. WTE_OK -> "Success"). */
/* Returns: string or NULL */
/*   err - error code */
WTAPIENTRY const char* wtrStrError(wt_error err);


/* Create a new Tomb virtual machine. */
/* Returns: error code (WTE_*) */
/*   p_profile  - pointer to the virtual machine will be returned here */
/*   ver        - desired VM version (set to 0 for most recent one) */
/*   flags      - OR'd WTRVMF_* constants */
/*   memdata    - optional memory array the VM can use as memory (if NULL */
/*                'alloc_func' will be called) */
/*   memsize    - desired size of the VM's memory array */
/*   ratio_css  - CallStackSize ratio for wtrContextCreate(); actual size */
/*                is computed with: */
/*                (stack_size*ratio_css)/(ratio_css+ratio_dss), */
/*                and will be rounded towards CSS */
/*   ratio_dss  - DataStackSize ratio for wtrContextCreate(); actual size */
/*                is computed the same way as CSS */
/*   alloc_ctx  - the context which will be passed to 'alloc_func' */
/*   alloc_func - memory allocation function */
WTAPIENTRY wt_error wtrCreateTombVM(wt_profile* const p_profile,
	wt_u32 ver, wt_u32 flags, void* memdata, wt_u32 memsize, wt_u8 ratio_css,
	wt_u8 ratio_dss, void* alloc_ctx, WT_PFN_ALLOC alloc_func);


/* Destroy the virtual machine context. */
/* Returns: error code (WTE_*) */
/*   profile - profile to destroy */
WTAPIENTRY wt_error wtrProfileDestroy(wt_profile profile);


/* Set operations used by the loader when an issue is found. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrProfileSetLoaderOps(wt_profile profile,
	void* const ops_ctx, const struct wtr_loaderops* const ops);


/* Set the opaque user pointer stored in the profile. */
/* Returns: error code (WTE_*) */
/*   profile - profile to change */
/*   ptr     - the new value of the per-profile user pointer */
WTAPIENTRY wt_error wtrProfileSetUserPointer(wt_profile profile, void* ptr);


/* Get the opaque user pointer stored in the profile. */
/* Returns: error code (WTE_*) */
/*   profile - profile to use */
/*   pptr    - store the value of the per-profile user pointer into */
WTAPIENTRY wt_error wtrProfileGetUserPointer(wt_profile profile, void** pptr);


/* Get the global ID of the exception. */
/* Returns: error code (WTE_*) */
/*   profile - search context */
/*   name    - name of the exception (zero terminated string) */
/*   p_exc   - write the global ID into this field */
WTAPIENTRY wt_error wtrProfileGetException(wt_profile profile,
	const char* const name, wt_exception* const p_exc);


/* Retrieve exception name from global exception ID. */
/* Returns: error code (WTE_*) */
/*   profile - search context */
/*   exc     - global exception ID */
/*   namenuf - buffer for name */
/*   bufsize - size of the buffer in characters */
WTAPIENTRY wt_error wtrProfileExceptionToString(wt_profile profile,
	wt_exception exc, char* namebuf, size_t bufsize);


/* Add a new exception into the context. */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   name    - name of the exception (zero terminated string) */
/*   out_exc - memory slot to store the exception ID of the new exception */
/*             (optional) */
WTAPIENTRY wt_error wtrProfileAddException(wt_profile profile,
	const char* const name, wt_exception* const out_exc);


/* Retrieve memory status of the virtual machine. This memory status may not accurate, */
/* because the programs allocate memory in larger blocks. */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   p_total - write the total size into this field */
/*   p_free  - write the free size into this field */
WTAPIENTRY wt_error wtrProfileMemoryStatus(wt_profile profile, wt_u64* p_total, wt_u64* p_free);


/* Add a new function into the context. */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   name    - name of the function (zero terminated string) */
/*   argc    - number of arguments */
/*   argv    - array of argument types (may NULL when 'argc' is zero) */
/*   retc    - number of return values */
/*   retv    - array of return value types (may NULL when 'retc' is zero) */
/*   func    - C function which will be executed when invoke the function */
/*   token   - a custom token to be registered with the function */
WTAPIENTRY wt_error wtrProfileAddFunction(wt_profile profile,
	const char* name, wt_u32 argc, const wt_u32* argv, wt_u32 retc,
	const wt_u32* retv, WT_PFN_EXTFUNC func, void* token);


/* Find and get a handle to a function. */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   p_func  - write the function handle into this field */
/*   name    - name of the function (zero terminated string) */
/*   argc    - number of arguments */
/*   argv    - array of argument types (may NULL when 'argc' is zero) */
/*   retc    - number of return values */
/*   retv    - array of return value types (may NULL when 'retc' is zero) */
/*   pure    - function is pure function or not */
WTAPIENTRY wt_error wtrProfileGetFunction(wt_profile profile,
	wt_function* p_func, const char* name, wt_u32 argc, const wt_u32* argv,
	wt_u32 retc, const wt_u32* retv, wt_bool pure);


/* Release a function handle which returned by wtrProfileGetFunction(). */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   func    - function handle to release */
WTAPIENTRY wt_error wtrProfileReleaseFunction(wt_profile profile,
	wt_function func);


/* Load a program into the given context. */
/* Returns: error code (WTE_*) */
/*   profile - virtual machine */
/*   program - write the program handle into this field */
/*   ctx     - context in initial state, used to execute the optional */
/*             program constructor */
/*   data    - WTB data */
/*   size    - size of 'data' in octets */
/*   flags   - additional flags (WTRLF_*) */
WTAPIENTRY wt_error wtrProfileLoadProgram(wt_profile profile,
	wt_program* program, struct wtr_context* const ctx, const void* data,
	size_t size, wt_u32 flags);


/* Release a program. */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   program - program to release */
WTAPIENTRY wt_error wtrProfileReleaseProgram(wt_profile profile,
	wt_program program, struct wtr_context* const ctx);


/* Set the opaque user pointer stored in the program. */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   program - program to change */
/*   ptr     - the new value of the per-program user pointer */
WTAPIENTRY wt_error wtrProgramSetUserPointer(wt_profile profile,
	wt_program program, void* ptr);


/* Get the opaque user pointer stored in the program. */
/* Returns: error code (WTE_*) */
/*   profile - context */
/*   program - program to use */
/*   pptr    - store the value of the per-program user pointer into */
WTAPIENTRY wt_error wtrProgramGetUserPointer(wt_profile profile,
	wt_program program, void** pptr);


/* Get the value of the specified metadata. If the metadata key is not */
/* found, NULL is returned. */
WTAPIENTRY const char* wtrProgramGetMetadata(wt_program program,
	const char* const key);


/* Create a new context that can run programs in the specified virtual */
/* machine. Context can be destroyed with wtrContextDestroy(), when it is */
/* in initial state. Newly created context will start in initial state. */
/* In initial state the context can be setup to execute a function or a */
/* special entry of a program. After the set up function returned, the */
/* context can/must be reset to initial state by calling */
/* wtrContextPopReturnValues(). */
/* Returns: error code (WTE_*) */
/*   p_context  - memory slot for the pointer to the created context */
/*   profile    - the virtual machine */
/*   stack_size - the context will have this amount of stack space (note */
/*                that, Tomb VM has separate data and call stack, and this */
/*                size will be split according to the ratio set in the VM); */
/*                minimal (128) and maximal (16MiB) stack size is enforced */
/*   sbuf_size  - string buffer to use in I/O calls (set to 0 to use the */
/*                default sbuf size; there is an internal minimal value */
/*                which will be enforced) */
/*   flags      - not used yet, must be zero */
WTAPIENTRY wt_error wtrContextCreate(struct wtr_context** const p_context,
	wt_profile profile, wt_u32 stack_size, wt_u32 sbuf_size, wt_u32 flags);


/* Destroy a context that is in initial state. When WTE_UCLEAN is returned, */
/* wtrContextCleanup() must be called, and the context may be destroyed */
/* after context returned to idle. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextDestroy(struct wtr_context* const context);


/* Start a cleanup in the context. After a successful call to this function */
/* wtrContextRun() must be called, until context does not return to idle. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextCleanup(struct wtr_context* const context);


/* Set the opaque user pointer stored in the context. This pointer is set */
/* to NULL when context is created. */
/* Returns: error code (WTE_*) */
/*   ptr - the new value of the per-context user pointer */
WTAPIENTRY wt_error wtrContextSetUserPointer(
	struct wtr_context* const context, void* ptr);


/* Get the opaque user pointer stored in the context. */
/* Returns: error code (WTE_*) */
/*   pptr - store the value of the per-program user pointer into */
WTAPIENTRY wt_error wtrContextGetUserPointer(
	struct wtr_context* const context, void** pptr);


/* Invoke the main{} of the specified program. Context must be in initial */
/* state. Sets up the context to executable state, see wtrContextRun(). */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextInvokeProgram(struct wtr_context* const context,
	wt_program program);


/* Invoke the specified function. Context must be in initial state. Number */
/* and types of the arguments must match with the functions arguments. Sets */
/* up the context to executable state, see wtrContextRun(). 'argv' can be */
/* NULL, iff 'argc' is zero. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextInvokeFunction(
	struct wtr_context* const context, wt_function func, wt_u32 argc,
	wt_value* argv);


/* Pop return values of the invoked entity if context state is RETURNED. */
/* Note that, some functions and main{} does not have return values, but */
/* this function must be called. Change the context state from RETURNED to */
/* INIT. Important: in case of returned object (i.e. string or seq), the */
/* memory area belong to these objects might be destroyed when the VM */
/* continues, thus it is advised to copy them if they needed for future */
/* use. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextPopReturnValues(
	struct wtr_context* const context, wt_u32* const retc, wt_u32 retmax,
	wt_value* retv);


/* Start or continue the execution of instructions in the Virtual Machine. */
/* Requires a set up context. Maximal executable instructions by a single */
/* call can be limited by 'quantum' -- if the VM executes the specified */
/* amount of intructions, it will return, thus the caller can schedule */
/* other activities. */
/* Returns: reason why the function returned (WTRR_*), further processing */
/* if the retuned value is mandatory to proper event handling */
WTAPIENTRY enum wtr_runresult wtrContextRun(struct wtr_context* const context,
	wt_u32 quantum);


/* Raise an exception in the context. Requires a set up context. After */
/* raising the exception, the context state will change to RUN, this call */
/* to wtrContextRun() is required to actually handle the exception. */
/* Acquire global exception ID by calling wtrProfileAddException() or */
/* wtrProfileGetException(). */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextRaiseException(
	struct wtr_context* const context, wt_exception exception_id);


/* Get the global ID of the unhandled exception by the context. Requires */
/* a context in EXCEPTION state. Context state will be changed from */
/* EXCEPTION to INIT. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextGetException(struct wtr_context* const context,
	wt_exception* const out_exception_id);


/* Get the parameters of the external function call. This function is */
/* needed only if the context is prohibited to call external functions or */
/* the external function yielded. Context state must be EXTCALL, and this */
/* state will not changed until wtrContextPushReturnValues() or */
/* wtrContextRaiseException() is called. Note that, if an object is in */
/* the arguments, its content must be copied if required for future use, */
/* otherwise they might be destroyed then the VM continues. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextGetExtCallParams(
	struct wtr_context* const context,
	struct wtr_extcallinfo** const p_callinfo, wt_value** const p_argv);


/* Get parameters of the I/O call. Requires a context in IOCALL state, */
/* and this state is preserved by this call. Context state will be changed */
/* by wtrContextPushReturnValues() or  wtrContextRaiseException(). Note */
/* that, if an object is in the arguments, its content must be copied if */
/* required for future use (however, for I/O calls this is unadvised), */
/* otherwise they might be destroyed then the VM continues. Unimplemented */
/* I/O calls should be handled by raising 'IOException'. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextGetIOCallParams(
	struct wtr_context* const context, enum wtr_iocall* const iocall,
	wt_u32* const argc, wt_u32 argmax, wt_value* argv);


/* Push the return values of an external function call or an I/O call. */
/* Important: this function must be called even when the call does not have */
/* actual return values, because this function is used to change the state */
/* of the context from EXTCALL or IOCALL to RUN. Number and types of the */
/* return values must match; see documentation for details. 'retv' can be */
/* NULL, iff 'retc' is zero. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextPushReturnValues(
	struct wtr_context* const context, wt_u32 retc, wt_value* retv);


/*----------------------------------------------------------------*/
#ifdef __cplusplus
}
#endif /* __cplusplus */
/*----------------------------------------------------------------*/

#endif /* WTRUNTIME_PUBLIC_H */
